# 0.3.0 Cleaned Up Log Of Changes

## Improved Segmentation Map Augmentation #302

Augmentation of Segmentation Maps is now faster and more memory efficient.
This required some breaking changes to `SegmentationMapOnImage`.
To adapt to the new version, the following steps should be sufficient for most
users:

* Rename all calls of `SegmentationMapOnImage` to `SegmentationMapsOnImage`
  (Map -> Maps).
* Rename all calls of `SegmentationMapsOnImage.get_arr_int()` to
  `SegmentationMapsOnImage.get_arr()`.
* Remove the argument `nb_classes` from all calls of `SegmentationMapsOnImage`.
* Remove the arguments `background_id` and `background_threshold` from all
  calls as these are no longer supported.
* Ensure that the input array to `SegmentationMapsOnImage` is always an
  int-like (int, uint or bool).
  Float arrays are no longer accepted.
* Adapt all calls `SegmentationMapsOnImage.draw()` and
  `SegmentationMapsOnImage.draw_on_image()`, as both of these now return a
  list of drawn images instead of a single array. (For a segmentation map
  array of shape `(H,W,C)` they return `C` drawn images. In most cases `C=1`,
  so simply call `draw()[0]` or `draw_on_image()[0]`.)
* Ensure that if `SegmentationMapsOnImage.arr` is accessed anywhere, the
  respective code can handle the new `int32` `(H,W,#maps)` array form.
  Previously it was `float32` and the channel-axis had the same size as the
  max class id (+1) that could appear in the map.

Changes:

- Changes to class `SegmentationMapOnImage`:
    - Renamed `SegmentationMapOnImage` to plural `SegmentationMapsOnImage`
      and deprecated the old name.
      This was changed due to the input array now being allowed to contain
      several channels, with each such channel containing one full segmentation
      map.
    - Changed `SegmentationMapsOnImage.__init__` to produce a deprecation
      warning for float arrays as `arr` argument.
    - **[breaking]** Changed `SegmentationMapsOnImage.__init__` to no longer
      accept `uint32` and larger itemsizes as `arr` argument, only `uint16`
      and below is accepted. For `int` the allowed maximum is `int32`.
    - Changed `SegmentationMapsOnImage.__init__` to always accept `(H,W,C)`
      `arr` arguments.
    - **[breaking]** Changed  `SegmentationMapsOnImage.arr` to always be
      `int32` `(H,W,#maps)` (previously: `float32` `(H,W,#nb_classes)`).
    - Deprecated `nb_classes` argument in `SegmentationMapsOnImage.__init__`.
      The argument is now ignored.
    - Added `SegmentationMapsOnImage.get_arr()`, which always returns a
      segmentation map array with similar dtype and number of dimensions as
      was originally input when creating a class instance.
    - Deprecated `SegmentationMapsOnImage.get_arr_int()`.
      The method is now an alias for `get_arr()`.
    - `SegmentationMapsOnImage.draw()`:
        - **[breaking]** Removed argument `return_foreground_mask` and
          corresponding optional output. To generate a foreground mask
          for the `c`-th segmentation map on a given image (usually `c=0`),
          use `segmentation_map.arr[:, :, c] != 0`, assuming that `0` is
          the integer index of your background class. 
        - **[breaking]** Changed output of drawn image to be a list of arrays
          instead of a single array (one per `C` in input array `(H,W,C)`).
        - Refactored to be a wrapper around
          `SegmentationMapsOnImage.draw_on_image()`.
        - The `size` argument may now be any of: A single `None` (keep shape),
          a single integer (use as height and width), a single float (relative
          change to shape) or a tuple of these values. ("shape" here denotes
          the value of the `.shape` attribute.)
    - `SegmentationMapsOnImage.draw_on_image()`:
        - **[breaking]** The argument `background_threshold` is now deprecated
          and ignored. Providing it will lead to a deprecation warning.
        - **[breaking]** Changed output of drawn image to be a list of arrays
          instead of a single array (one per `C` in input array `(H,W,C)`).
    - Changed `SegmentationMapsOnImage.resize()` to use nearest neighbour
      interpolation by default.
    - **[rarely breaking]** Changed `SegmentationMapsOnImage.copy()` to create
      a shallow copy instead of being an alias for `deepcopy()`.
    - Added optional arguments `arr` and `shape` to
      `SegmentationMapsOnImage.copy()`.
    - Added optional arguments `arr` and `shape` to
      `SegmentationMapsOnImage.deepcopy()`.
    - Refactored `SegmentationMapsOnImage.pad()`,
      `SegmentationMapsOnImage.pad_to_aspect_ratio()` and
      `SegmentationMapsOnImage.resize()` to generate new object instances via
      `SegmentationMapsOnImage.deepcopy()`.
    - **[rarely breaking]** Renamed `SegmentationMapsOnImage.input_was` to
      `SegmentationMapsOnImage._input_was`.
    - **[rarely breaking]** Changed `SegmentationMapsOnImage._input_was` to
      always save `(input array dtype, input array ndim)` instead of mixtures
      of strings/ints that varied by dtype kind.
    - **[rarely breaking]** Restrict `shape` argument in
      `SegmentationMapsOnImage.__init__` to tuples instead of accepting all
      iterables.
    - **[breaking]** Removed `SegmentationMapsOnImage.to_heatmaps()` as the
      new segmentation map class is too different to sustain the old heatmap
      conversion methods.
    - **[breaking]** Removed `SegmentationMapsOnImage.from_heatmaps()` as the
      new segmentation map class is too different to sustain the old heatmap
      conversion methods.
- Changes to class `Augmenter`:
    - **[breaking]** Automatic segmentation map normalization from arrays or
      lists of arrays now expects a single `(N,H,W,C)` array (before:
      `(N,H,W)`) or a list of `(H,W,C)` arrays (before: `(H,W)`).
      This affects valid segmentation map inputs for `Augmenter.augment()`
      and its alias `Augmenter.__call__()`,
      `imgaug.augmentables.batches.UnnormalizedBatch()` and
      `imgaug.augmentables.normalization.normalize_segmentation_maps()`.
    - Added `Augmenter._augment_segmentation_maps()`.
    - Changed `Augmenter.augment_segmentation_maps()` to no longer be a
    wrapper around `Augmenter.augment_heatmaps()` and instead call
    `Augmenter._augment_segmentation_maps()`.
- Added special segmentation map handling to all augmenters that modified
  segmentation maps
  (`Sequential`, `SomeOf`, `Sometimes`, `WithChannels`,
   `Lambda`, `AssertLambda`, `AssertShape`,
   `Alpha`, `AlphaElementwise`, `WithColorspace`, `Fliplr`, `Flipud`, `Affine`,
   `AffineCv2`, `PiecewiseAffine`, `PerspectiveTransform`, `ElasticTransformation`,
   `Rot90`, `Resize`, `CropAndPad`, `PadToFixedSize`, `CropToFixedSize`,
   `KeepSizeByResize`).
   - **[rarely breaking]** This changes the order of arguments in
     `Lambda.__init__()`, `AssertLambda.__init__()`, `AssertShape.__init__()`
     and hence breaks if one relied on that order.

## New RNG handling #375

* Adapted library to automatically use the new `numpy.random` classes of
  numpy 1.17 -- if they are available. If they are not available (i.e. numpy
  version is <=1.16), the library automatically falls back to the old
  interface (i.e. `numpy.random.RandomState`).
* Added module `imgaug.random`.
  * Added class `imgaug.random.RNG`. This is now the preferred way to represent
    RNG states (previously: `numpy.random.RandomState`). Instantiate it
    via e.g. `RNG(1052912236)`, where `1052912236` is a seed.
  * Added `imgaug.random.supports_new_rng_style()`.
  * Added `imgaug.random.get_global_rng()`.
  * Added `imgaug.random.seed()`.
  * Added `imgaug.random.normalize_generator()`.
  * Added `imgaug.random.normalize_generator_()`.
  * Added `imgaug.random.convert_seed_to_generator()`.
  * Added `imgaug.random.convert_seed_sequence_to_generator()`.
  * Added `imgaug.random.create_pseudo_random_generator_()`.
  * Added `imgaug.random.create_fully_random_generator()`.
  * Added `imgaug.random.generate_seed_()`.
  * Added `imgaug.random.generate_seeds_()`.
  * Added `imgaug.random.copy_generator()`.
  * Added `imgaug.random.copy_generator_unless_global_generator()`.
  * Added `imgaug.random.reset_generator_cache_()`.
  * Added `imgaug.random.derive_generator_()`.
  * Added `imgaug.random.derive_generators_()`.
  * Added `imgaug.random.get_generator_state()`.
  * Added `imgaug.random.set_generator_state_()`.
  * Added `imgaug.random.is_generator_equal_to()`.
  * Added `imgaug.random.advance_generator_()`.
  * Added `imgaug.random.polyfill_integers()`.
  * Added `imgaug.random.polyfill_random()`.
* Refactored all arguments related to random state handling to also accept
  `imgaug.random.RNG`, as well as the new numpy random classes. This
  particularly affects `imgaug.augmenters.meta.Augmenter` and
  `imgaug.parameters.StochasticParameter` (argument `random_state` for both).
* Marked old RNG related functions in `imgaug.imgaug` as deprecated.
  They will now produce warnings and redirect towards corresponding functions
  in `imgaug.random`. This does not yet affect `imgaug.imgaug.seed()`.
  It does affect the functions listed below.
  * `imgaug.imgaug.normalize_random_state()`.
  * `imgaug.imgaug.current_random_state()`.
  * `imgaug.imgaug.new_random_state()`.
  * `imgaug.imgaug.dummy_random_state()`.
  * `imgaug.imgaug.copy_random_state()`.
  * `imgaug.imgaug.derive_random_state()`.
  * `imgaug.imgaug.derive_random_states()`.
  * `imgaug.imgaug.forward_random_state()`.
* [rarely breaking] Removed `imgaug.imgaug.CURRENT_RANDOM_STATE`.
  Use `imgaug.random.get_global_rng()` instead.
* [rarely breaking] Removed `imgaug.imgaug.SEED_MIN_VALUE`.
  Use `imgaug.random.SEED_MIN_VALUE` instead or sample seeds via
  `imgaug.random.generate_seeds_()`.
* [rarely breaking] Removed `imgaug.imgaug.SEED_MAX_VALUE`.
  Use `imgaug.random.SEED_MAX_VALUE` instead or sample seeds via
  `imgaug.random.generate_seeds_()`.
* Optimized RNG handling throughout all augmenters to minimize the number of
  RNG copies. RNGs are now re-used as often as possible. This improves
  performance, but has the disadvantage that adding images to a batch will now
  often affect the samples of the other images in the same batch. E.g.
  previously for a batch of images `A,B,C` and seed `1`, the samples of `A,B,C`
  would remain unchanged if the batch was changed to `A,B,C,D` (provided the
  seed stayed the same). Now, if `D` is added the samples of `A,B,C` may
  change.
* [breaking] The above listed changes will lead to different values being
  sampled for the same seeds (compared to past versions of the library).
* [breaking] The seed for `imgaug`'s global random number generator is now
  sampled from numpy's default random number generator. That means, that every
  run of a program using `imgaug` will by default use a different seed and
  hence result in different samples. Previously, a fixed seed was used,
  resulting in the same samples for each run (unless the seed was manually
  changed to a fixed one). It also means that seeding numpy will automatically
  also seed imgaug (not guarantueed that this behaviour will be kept in
  future releases). The change from fixed to random seed was done, because the
  old (fixed) behaviour didn't match the common practice (and especially not
  numpy's standard behaviour) and hence led to confusion. #408


## Adaptations to numpy 1.17

* [rarely breaking] Deactivated support for `int64` in
  `imgaug.dtypes.clip_()`. This is due to numpy 1.17 turning `int64` to
  `float64` in `numpy.clip()` (possible that this happened in some way
  before 1.17 too). #302
* [rarely breaking] Changed `imgaug.dtypes.clip()` to never clip `int32`
  in-place, as `numpy.clip()` turns it into `float64` since 1.17 (possible
  that this happend in some way before 1.17 too).
* [rarely breaking] Deactivated support for `int64` in
  `ReplaceElementwise`. See `clip` issue above. #302
* [rarely breaking] Changed `parameters.DiscreteUniform` to always return
  arrays of dtype `int32`. Previously it would automatically return
  `int64`. #302
* [rarely breaking] Changed `parameters.Deterministic` to always return
  `int32` for integers and always `float32` for floats. #302
* [rarely breaking] Changed `parameters.Choice` to limit integer
  dtypes to `int32` or lower, uints to `uint32` or lower and floats
  to `float32` or lower. #302
* [rarely breaking] Changed `parameters.Binomial` and `parameters.Poisson`
  to always return `int32`. #302
* [rarely breaking] Changed `parameters.Normal`,
  `parameters.TruncatedNormal`, `parameters.Laplace`,
  `parameters.ChiSquare`, `parameters.Weibull`, `parameters.Uniform` and
  `parameters.Beta` to always return `float32`. #302
* [rarely breaking] Changed `augmenters.arithmetic.Add`,
  `augmenters.arithmetic.AddElementwise`, `augmenters.arithmetic.Multiply`
  and `augmenters.arithmetic.MultiplyElementwise` to no longer internally
  increase itemsize of dtypes by a factor of 2 for
  dtypes `uint16`, `int8` and `uint16`. For `Multiply*` this also
  covers `float16` and `float32`. This protects against crashes due to
  clipping `int64` or `uint64` data. In rare cases this can lead to
  overflows if `image + random samples` or `image * random samples`
  exceeds the value range of `int32` or `uint32`. This change may affect
  various other augmenters that are wrappers around the mentioned ones,
  e.g. `AdditiveGaussianNoise`. #302
* [rarely breaking] Decreased support of dtypes `uint16`, `int8`,
  `int16`, `float16`, `float32` and `bool` in `augmenters.arithmetic.Add`,
  `AddElementwise`, `Multiply` and `MultiplyElementwise` from "yes" to
  "limited". #302
* [rarely breaking] Decreased support of dtype `int64` in
  `augmenters.arithmetic.ReplaceElementwise` from "yes" to "no". This also
  affects all `*Noise` augmenters (e.g. `AdditiveGaussianNoise`,
  `ImpulseNoise`), all `Dropout` augmenters, all `Salt` augmenters and
  all `Pepper` augmenters. #302
* [rarely breaking] Changed `augmenters.contrast.adjust_contrast_log`
  and thereby `LogContrast` to no longer support dtypes `uint32`, `uint64`,
  `int32` and `int64`. #302


## New Augmenters


* Added `augmenters.edges.Canny`, which applies canny edge detection with alpha
  blending and random coloring to images. #316
* Added `augmenters.pooling.AveragePooling`. #317
* Added `augmenters.pooling.MaxPooling`. #317
* Added `augmenters.pooling.MinPooling`. #317
* Added `augmenters.pooling.MedianPooling`. #317
* Added `augmenters.color.AddToHue`, a shortcut for
  `AddToHueAndSaturation(value_hue=...)`. #319
* Added `augmenters.color.AddToSaturation`, a shortcut for
  `AddToHueAndSaturation(value_saturation=...)`. #319
* Added `augmenters.color.WithHueAndSaturation`. #319
* Added `augmenters.color.MultiplyHueAndSaturation`. #319
* Added `augmenters.color.MultiplyHue`. #319
* Added `augmenters.color.MultiplySaturation`. #319
* Added `augmenters.color.KMeansColorQuantization` and corresponding
  `augmenters.color.quantize_colors_kmeans()`. Both deal with quantizing
  similar colors using k-Means clustering. #347
    * Added a check script for `KMeansColorQuantization` under
      `checks/check_kmeans_color_quantization.py`. #347
* Added `augmenters.color.UniformColorQuantization` and corresponding
  `augmenters.color.quantize_colors_uniform()`. Both deal with quantizing
  similar colors using k-Means clustering. #347
* Added `imgaug.augmenters.segmentation.Voronoi`. An augmenter that converts
  an image to a voronoi image.  #348
* Added `imgaug.augmenters.segmentation.UniformVoronoi`, a shortcut for
  `Voronoi(UniformPointsSamper)`. #348
* Added `imgaug.augmenters.segmentation.RegularGridVoronoi`, a shortcut for
  `Voronoi(DropoutPointsSampler(RegularGridPointsSampler))`. #348
* Added `imgaug.augmenters.segmentation.RelativeRegularGridVoronoi`, a shortcut
  for `Voronoi(DropoutPointsSampler(RelativeRegularGridPointsSampler))`. #348


## New Modules

* Added module `imgaug.augmenters.edges`. #316
* Added module `imgaug.augmenters.pooling`. #317
* Added module `imgaug.validation`. The module is intended for functions
  related to the validation of input arguments. #413


## output_buffer_size

* Added argument `output_buffer_size` to `multicore.Pool.imap_batches()`
  and `multicore.Pool.imap_batches_unordered()` to control the maximum number
  of batches in the background augmentation pipeline (allows to limit
  maximum RAM demands). #305
* Changed default `output_buffer_size` in `Augmenter.augment_batches()` from
  "unlimited" to `10*C`, where `C` is the number of logical CPU cores. #417


## Other New Classes

* Added interface `augmenters.edges.BinaryImageColorizerIf`, which
  contains the interface for classes used to convert binary images to RGB
  images. #316
* Added `augmenters.pooling._AbstractPoolingBase`. #317
* Added `augmenters.edges.RandomColorsBinaryImageColorizer`, which
  converts binary images to RGB images by sampling uniformly RGB colors for
  `True` and `False` values. #316
* Added `augmenters.color._AbstractColorQuantization`. #347
* Added `imgaug.augmenters.segmentation.PointsSamplerIf`. An interface for
  classes used for sampling (usually random) coordinate arrays on images. #348
* Added `imgaug.augmenters.segmentation.RegularGridPointsSampler`. A class
  used to generate regular grids of `rows x columns` points on images. #348
* Added `imgaug.augmenters.segmentation.RelativeRegularGridPointsSampler`.
  Similar to `RegularGridPointsSampler`, but number of rows/columns is set
  as fractions of image sizes, leading to more rows/columns for larger
  images. #348
* Added `imgaug.augmenters.segmentation.DropoutPointsSampler`. A class
  used to randomly drop `p` percent of all coordinates sampled by another
  another points sampler. #348
* Added `imgaug.augmenters.segmentation.UniformPointsSampler`. A class used
  to sample `N` points on each image with y-/x-coordinates uniformly sampled
  using the corresponding image height/width. #348 
* Added `imgaug.augmenters.segmentation.SubsamplingPointsSampler`. A class
  that ensures that another points sampler does not produce more than
  `N` points by subsampling a random subset of the produced points if `N`
  is exceeded. #348
* Added `imgaug.testutils.ArgCopyingMagicMock`. #413


## Other New Functions

* Added `imgaug.is_np_scalar()`, analogous to `imgaug.is_np_array()`. #366
* Added `dtypes.normalize_dtypes()`. #366
* Added `dtypes.normalize_dtype()`. #366
* Added `dtypes.change_dtypes_()`. #366
* Added `dtypes.change_dtype_()`. #366
* Added `dtypes.increase_itemsize_of_dtype()`. #366
* Added `imgaug.warn()` function. #367
* Added `imgaug.min_pool()`. #369
* Added `imgaug.median_pool()`. #369
* Added `imgaug.compute_paddings_to_reach_multiples_of()`. #369
* Added `imgaug.pad_to_multiples_of()`. #369
* Added `imgaug.imgaug.normalize_random_state()`. #348
* Added `imgaug.augmenters.segmentation._ensure_image_max_size()`. #348
* Added `imgaug.augmenters.segmentation._verify_sample_points_images()`. #348
* Added `imgaug.augmenters.segmentation.segment_voronoi()`, a function that
  converts an image into a voronoi image, i.e. averages the colors within
  voronoi cells placed on the image. #348
* Added `_match_pixels_with_voronoi_cells()`. #348
* Added `_generate_pixel_coords()`. #348
* Added `_compute_avg_segment_colors()`. #348
* Added `_render_segments()`. #348
* Added `augmentables.utils.copy_augmentables`. #410
* Added `augmenters.flip.fliplr()`. #385
* Added `augmenters.flip.flipud()`. #385
* Added `augmenters.color.change_colorspace_()`. #409
* Added `augmenters.color.change_colorspace_batch_()`. #409
* Added `augmenters.arithmetic.add_scalar()`. #411
* Added `augmenters.arithmetic.add_elementwise()`. #411
* Added `augmenters.arithmetic.replace_elementwise_()`. #411
* Added `augmenters.arithmetic.compress_jpg()`. #411
* Added `validation.convert_iterable_to_string_of_types()`. #413
* Added `validation.is_iterable_of()`. #413
* Added `validation.assert_is_iterable_of()`. #413


## Other New Constants

* Added to `imgaug.augmenters.color` the constants `CSPACE_RGB`, 
  `CSPACE_BGR`, `CSPACE_GRAY`, `CSPACE_CIE`, `CSPACE_YCrCb`, `CSPACE_HSV`,
  `CSPACE_HLS`, `CSPACE_Lab`, `CSPACE_Luv`, `CSPACE_YUV`, `CSPACE_ALL`. #409


## Other New Arguments

* [rarely breaking] Added a `pad_mode` argument to `imgaug.pool()`,
  `imgaug.avg_pool()`, `imgaug.max_pool()`, `imgaug.min_pool()` and
  `imgaug.median_pool()`. This breaks code relying on the order of the
  functions arguments. #369
* Added to argument `size` of `Resize` the optional keys `short-side` and
  `longer-side`. This adds the ability to resize the shorter and longer sides
  of images (instead of only height/width). #349
* [rarely breaking] Added `value_hue` and `value_saturation` arguments,
  which allow to set individual parameters for hue and saturation
  instead of having to use one parameter for both (they may not be set
  if `value` is already set).
  This changes the order of arguments of the augmenter and code that relied
  on that order will now break.
  This also changes the output of
  `AddToHueAndSaturation.get_parameters()`. #319
* [rarely breaking] Added argument `polygon_recoverer` to
  `augmenters.geometric.PerspectiveTransform`. This changes the order of
  arguments of the augmenter and code that relied on that order will now
  break. #338
* Added attribute `from_colorspace` to `AddToHueAndSaturation`. #409


## Other Removed Concepts

* [rarely breaking] Removed `dtypes.get_minimal_dtype_for_values()`. The
  function was not used anywhere in the library. #366
* [rarely breaking] Removed `dtypes.get_minimal_dtype_by_value_range()`. The
  function was not used anywhere in the library. #366
* [rarely breaking] Removed `Affine.VALID_DTYPES_CV2_ORDER_0`. #407
* [rarely breaking] Removed `Affine.VALID_DTYPES_CV2_ORDER_NOT_0`. #407
* [rarely breaking] Removed `Affine.order_map_skimage_cv2`. #407
* [rarely breaking] Removed `Affine.mode_map_skimage_cv2`. #407
* [rarely breaking] Removed attributes `colorspace_changer` and
  `colorspace_changer_inv` from `AddToHueAndSaturation`. #409
* [rarely breaking] Removed class constant `ALLOW_DTYPES_CUSTOM_MINMAX`
  from `Invert`. #411
* [rarely breaking] Removed attribute `dtype_kind_to_invert_func` from
  `Invert`. #411
* [rarely breaking] Removed attribute `maximum_quality` from
  `JpegCompression`. #411
* [rarely breaking] Removed attribute `minimum_quality` from
  `JpegCompression`. #411
* [rarely breaking] Removed argument `affects` from
  `dtypes.promote_array_dtypes_()` as it was unnecessary and not used anywhere
  in the library. #366
* [rarely breaking] Removed method
  `ElasticTransformation.generate_shift_maps()`. Use
  `ElasticTransformation._generate_shift_maps()` instead. #413
* [rarely breaking] Removed method `ElasticTransformation.map_coordinates()`.
  Use `ElasticTransformation._map_coordinates()` instead. #413


## Performance

* Replaced all calls of `imgaug.do_assert` with ordinary `assert`
  statements. This is a bit less secure, but should overall improve
  performance. #387
* Improved performance of `augmenters.flip.Fliplr`. #385
* Improved performance of `augmenters.flip.Flipud`. #385


## Deprecation

* Marked the following functions as deprecated (#398):
  * `imgaug.augmenters.meta.clip_augmented_image_`
  * `imgaug.augmenters.meta.clip_augmented_image`
  * `imgaug.augmenters.meta.clip_augmented_images_`
  * `imgaug.augmenters.meta.clip_augmented_images`
* Marked `imgaug.augmenters.arithmetic.ContrastNormalization` as deprecated.
  Use `imgaug.augmenters.contrast.LinearContrast` instead. #396
* Marked argument `X` of `imgaug.augmentables.kps.compute_geometric_median()`
  as deprecated. Use argument `points` instead. #402
* Marked `cval` in `imgaug.pool()`, `imgaug.avg_pool()` and `imgaug.max_pool()`
  as deprecated. Use `pad_cval` instead. #369


## Refactorings

* Refactored `augmenters.arithmetic` to improve code quality and
  docstrings. #328
* Refactored `augmenters.segmentation` to improve code quality and
  docstrings. #334
* Refactored `augmenters/convolutional.py` to improve code quality and
  docstrings. #335
* Refactored `augmenters/weather.py` to improve code quality and
  docstrings. #336
* Refactored `multicore` to improve code quality and
  docstrings. #367
  * Improved error messages in `multicore`.
* Refactored `imgaug.pool()` to use `imgaug.pad()` for image padding. #369
* Refactored `augmenters.pooling.MedianPooling` to use
  `imgaug.median_pool()`. #369
* Refactored `augmenters.pooling.MinPooling` to use `imgaug.min_pool()`. #369
* Improved code style and documentation of (#389, #402):
    * `imgaug.augmentables.bbs`.
    * `imgaug.augmentables.heatmaps`.
    * `imgaug.augmentables.kps`.
    * `imgaug.augmentables.lines`.
    * `imgaug.augmentables.normalization`.
    * `imgaug.augmentables.polys`.
    * `imgaug.augmentables.segmaps`.
    * `imgaug.augmentables.utils`.
    * `imgaug.imgaug`.
    * `imgaug.parameters`.
    * `imgaug.augmenters.weather`.
    * `imgaug.augmenters.size`.
    * `imgaug.augmenters.segmentation`.
    * `imgaug.augmenters.meta`.
    * `imgaug.augmenters.geometric`.
    * `imgaug.augmenters.flip`.
    * `imgaug.augmenters.contrast`.
    * `imgaug.augmenters.blur`.
    * `imgaug.augmenters.blend`.
    * `imgaug.augmenters.weather`.
* Refactored all calls of `warnings.warn()` to use `imgaug.imgaug.warn()
  instead. #401
* [rarely breaking] Refactored most of the augmenters from functions to
  classes. Previously, some augmenters were functions that returned an
  instance of another augmenter (with adjusted hyperparameters) when being
  called. Aside from a few corner cases, these have been switched to classes
  inheriting from the augmenters that were previously returned. This should
  make some outputs less confusing (e.g. `print(A())` does no longer lead to
  class `B` being printed). All arguments stayed the same and this is not
  expected to affect any user code negatively. The augmenters listed below are
  affected by this change. #396
    * `imgaug.augmenters.arithmetic.AdditiveGaussianNoise`
    * `imgaug.augmenters.arithmetic.AdditiveLaplaceNoise`
    * `imgaug.augmenters.arithmetic.AdditivePoissonNoise`
    * `imgaug.augmenters.arithmetic.Dropout`
    * `imgaug.augmenters.arithmetic.CoarseDropout`
    * `imgaug.augmenters.arithmetic.ImpulseNoise`
    * `imgaug.augmenters.arithmetic.SaltAndPepper`
    * `imgaug.augmenters.arithmetic.CoarseSaltAndPepper`
    * `imgaug.augmenters.arithmetic.Salt`
    * `imgaug.augmenters.arithmetic.CoarseSalt`
    * `imgaug.augmenters.arithmetic.Pepper`
    * `imgaug.augmenters.arithmetic.CoarsePepper`
    * `imgaug.augmenters.blend.SimplexNoiseAlpha`
    * `imgaug.augmenters.blend.FrequencyNoiseAlpha`
    * `imgaug.augmenters.blur.MotionBlur`
    * `imgaug.augmenters.contrast.MultiplyHueAndSaturation`
    * `imgaug.augmenters.contrast.MultiplyHue`
    * `imgaug.augmenters.contrast.MultiplySaturation`
    * `imgaug.augmenters.contrast.AddToHue`
    * `imgaug.augmenters.contrast.AddToSaturation`
    * `imgaug.augmenters.contrast.Grayscale`
    * `imgaug.augmenters.contrast.GammaContrast`
    * `imgaug.augmenters.contrast.SigmoidContrast`
    * `imgaug.augmenters.contrast.LogContrast`
    * `imgaug.augmenters.contrast.LinearContrast`
    * `imgaug.augmenters.convolutional.Sharpen`
    * `imgaug.augmenters.convolutional.Emboss`
    * `imgaug.augmenters.convolutional.EdgeDetect`
    * `imgaug.augmenters.convolutional.DirectedEdgeDetect`
    * `imgaug.augmenters.meta.OneOf`
    * `imgaug.augmenters.meta.AssertLambda`
    * `imgaug.augmenters.meta.AssertShape`
    * `imgaug.augmenters.size.Pad`
    * `imgaug.augmenters.size.Crop`
    * `imgaug.augmenters.weather.Clouds`
    * `imgaug.augmenters.weather.Fog`
    * `imgaug.augmenters.weather.Snowflakes`
* Refactored `Affine` to improve code quality and minimize code
  duplication. #407
* Refactored `CropAndPad` to improve code quality and minimize code
  duplication. #407
* Refactored module `size` to decrease code duplication between different
  augmenters. #407
* Moved matrix generation logic of augmenters in module `convolutional`
  to classes, one per augmenter (i.e. one per category of convolutional
  matrix). This should avoid errors related to pickling of functions. #407
* Refactored color augmenters to use `change_colorspace_()` and
  `change_colorspace_batch_()`. #409
* Refactored `Alpha` to decrease code duplication. #410
* Refactored `AlphaElementwise` to decrease code duplication. #410
* Refactored `Add` to use `imgaug.augmenters.arithmetic.add_scalar()`. #411
* Refactored `AddElementwise` to use
  `imgaug.augmenters.arithmetic.add_elementwise()`. #411
* Refactored `ReplaceElementwise` to use
  `imgaug.augmenters.arithmetic.replace_elementwise_()`. #411
* Refactored `JpegCompression` to use
  `imgaug.augmenters.arithmetic.compress_jpg()`. #411
* Refactored `AddToHueAndSaturation` to decrease code duplication and improve
  code quality. #319
* Refactored `Affine` to improve code quality and decrease code
  duplication. #413
* Refactored `PiecewiseAffine` to improve code quality and decrease code
  duplication. #413
* Refactored `PerspectiveTransform` to improve code quality and decrease code
  duplication. #413
* Refactored `ElasticTransformation` to improve code quality and decrease code
  duplication. #413
* Refactored `Rot90` to improve code quality and decrease code
  duplication. #413
* Refactored `Augmenter.augment_images()`, `Augmenter.augment_heatmaps()`,
  `Augmenter.augment_segmentation_maps()`, `Augmenter.augment_polygons()`,
  `Augmenter.augment_line_strings()` and `Augmenter._augment_coord_augables()`
  to improve code quality and remove redundancies. #413
* Refactored `imgaug.imgaug.imresize_single_image()`. #413
* Refactored `Sequential` to reduce code duplication. #413
* Refactored `SomeOf` to improve code quality. #413
* Refactored `Sometimes` to reduce code duplication. #413
* Refactored `AssertShape` to reduce code duplication. #413
* Refactored `ChannelShuffle` to improve code quality. #413
* Refactored `dtypes.py` to improve code quality. #366
* Refactored `dtypes.promote_array_dtypes_()` to use
  `dtypes.change_dtypes_()`. #366
* Refactored `dtypes.get_minimal_dtype()` to use
  `dtypes.increase_itemsize_of_dtype()`. #366
* Renamed `dtypes.restore_dtypes_()` to `dtypes.change_dtypes_()`. (Old name
  still exists too and redirects to new name. Not yet marked as
  deprecated.) #366


## Other Changes

* [rarely breaking] Changed colorspace transformations throughout the
  library to fail if the input image does not have three channels. #409
* Changed colorspace transformations throughout the library to also
  support `YUV` colorspace. #409
* [rarely breaking] Changed `AlphaElementwise` to verify for keypoint
  and line string augmentation that the number of coordinates before/after 
  augmentation does not change. Previously this was allowed. This also
  affects `SigmoidNoiseAlpha` and `FrequenceNoiseAlpha`. #410
* [rarely breaking] Changed `AlphaElementwise` to use for keypoint,
  line string and bounding box augmentation a pointwise approach, where
  per coordinate a decision is made whether the new coordinate from the
  first branch's (augmented) results or the second branch's (augmented)
  results are used. The decision is based on the average alpha mask value
  at the xy-location of the coordinate. For polygons, the old mode is
  still used where either all coordinates from the first branch's results
  or the second branch's results are used. This also affects 
  `SigmoidNoiseAlpha` and `FrequenceNoiseAlpha`. #410
* Improved error messages of `dtypes.restore_dtypes_()`. #366 


## Other Minor Changes

* Increased `max_distance` thresholds for `almost_equals()`,
  `exterior_almost_equals()` and `coords_almost_equals()` in `Polygon` and
  `LineString` from `1e-6` to `1e-4`. This should fix false-negative problems
  related to float inaccuracies.
* Changed `_ConcavePolygonRecoverer` to not search for segment intersection
  points in polygons with very large absolute coordinate values.
  This prevents rare errors due to floating point inaccuracies. #338
* Changed `_ConcavePolygonRecoverer` to raise warnings instead of throwing
  exceptions when the underlying search for segment intersection points
  crashes. #338
* Added check to `dtypes.gate_dtypes()` verifying that arguments `allowed`
  and `disallowed` have no intersection. #346
* Changed `dtypes.get_value_range_of_dtype()` to return a float as the center
  value of `uint` dtypes. #366
* Changed `multicore.Pool` to produce a warning if it cannot find or call the
  function `multiprocessing.cpu_count()` instead of silently failing.
  (In both cases it falls back to a default value.) #367
* Changed the default `pad_mode` of `avg_pool` from `constant` (`cval=128`)
  to `reflect`. #369
* Changed the default `pad_mode` of `max_pool` from `constant` (`cval=0`)
  to `edge`. #369
* Changed the default `pad_mode` of `min_pool` from `constant` (`cval=255`)
  to `edge`. #369
* Changed the default `pad_mode` of `median_pool` from `constant`
  (`cval=128`) to `reflect`. #369
* Changed `imgaug.imgaug.pad` to automatically clip the `cval` argument
  to the value range of the array to be padded. #407
* [rarely breaking] Changed `KeypointsOnImage.from_keypoints_image()` to
  return `(x+0.5, y+0.5)` instead of `(x, y)` where `(x, y)` denotes the
  coordinates of the pixel in which a maximum was found. This change matches
  the standard that all pixels are given with subpixel accuracy and therefore
  any whole pixel with a maximum should denote the coordinates of that
  pixel's center. #413
* Removed image-channel check for cv2-based warp in `Affine`. Images with any
  channel number can now be warped using the cv2 backend (previously: only
  `<=4`, others would be warped via skimage). #381
* [rarely breaking] The `value` parameter in
  `augmenters.color.AddToHueAndSaturation` is now interpreted by the
  augmenter to return first the hue and then the saturation value to add,
  instead of the other way round. (This isn't expected to affect
  anybody.) #319
* Added output `from_colorspace` to
  `AddToHueAndSaturation.get_parameters()`. #409
* Added dtype gating to `dtypes.clip_()`.
* Changed all dtype-related functions to accept also dtypes given as
  string names, numpy arrays, numpy scalars or dtype functions. #366
* Changed `dtypes.restore_dtypes_()` so that if `images` is a list of length
  `N` and `dtypes` is a list of length `M` and `N!=M`, the function now raises
  an `AssertionError`. #366
* Changed `dtypes.restore_dtypes_()` to ignore the argument `round` if the
  input array does not have a float dtype. #366
* Removed restrictions of `value` parameter in `AddElementwise`.
  The value range is now no longer limited to `[-255, 255]` and floats
  are now allowed. #411


## New Scripts

* Added a check script for `UniformColorQuantization` in
  `checks/check_uniform_color_quantization.py`. #347
* Added a check script for `Voronoi` in `checks/check_voronoi.py`. #348
* Added a check script for flip performance measurments under
  `checks/check_flip_performance.py`. #385


## Other

* Added the library to `conda-forge` so it can now be installed via
  `conda install imgaug` (provided the conda-forge channel was added
  before that). #320 #339
* Changed dependency `opencv-python` to `opencv-python-headless`.
  This should improve support for some system without GUIs.
* Added dependency `pytest-subtests` for the library's unittests. #366
* Improved the docstrings of most augmenters and added code examples. #302
* Added error messages to `assert` statements throughout the library. #387
* Removed the requirement to implement `_augment_keypoints()` and
  `_augment_heatmaps()` in augmenters. The methods now default to doing
  nothing. Also removed all such noop-implementations of these methods from
  all augmenters. #380
* Increased minimum version requirement for `scikit-image` to
  `0.14.2`. #377, #399
* Renamed `imgaug/external/poly_point_isect.py` to
  `imgaug/external/poly_point_isect_py3.py.bak`.
  The file is in the library only for completeness and contains python3 syntax.
  `poly_point_isect_py2py3.py` is actually used.


## Fixes
 
* Fixed an issue with `Polygon.clip_out_of_image()`,
  which would lead to exceptions if a polygon had overlap with an image,
  but not a single one of its points was inside that image plane. 
* Fixed `multicore` methods falsely not accepting
  `augmentables.batches.UnnormalizedBatch`.
* `Rot90` now uses subpixel-based coordinate remapping.
  I.e. any coordinate `(x, y)` will be mapped to `(H-y, x)` for a rotation by
  90deg.
  Previously, an integer-based remapping to `(H-y-1, x)` was used.
  Coordinates are e.g. used by keypoints, bounding boxes or polygons.
* `augmenters.arithmetic.Invert`
    * [rarely breaking] If `min_value` and/or `max_value` arguments were
      set, `uint64` is no longer a valid input array dtype for `Invert`.
      This is due to a conversion to `float64` resulting in loss of resolution.
    * Fixed `Invert` in rare cases restoring dtypes improperly.
* Fixed `dtypes.gate_dtypes()` crashing if the input was one or more numpy
  scalars instead of numpy arrays or dtypes.
* Fixed `augmenters.geometric.PerspectiveTransform` producing invalid
  polygons (more often with higher `scale` values). #338
* Fixed errors caused by `external/poly_point_isect_py2py3.py` related to
  floating point inaccuracies (changed an epsilon from `1e-10` to `1e-4`,
  rounded some floats). #338
* Fixed `Superpixels` breaking when a sampled `n_segments` was `<=0`.
  `n_segments` is now treated as `1` in these cases.
* Fixed `ReplaceElementwise` both allowing and disallowing dtype `int64`. #346
* Fixed `BoundingBox.deepcopy()` creating only shallow copies of labels. #356
* Fixed `dtypes.change_dtypes_()` #366
    * Fixed argument `round` being ignored if input images were a list.
    * Fixed failure if input images were a list and dtypes a single numpy
      dtype function.
* Fixed `dtypes.get_minimal_dtype()` failing if argument `arrays` contained
  not *exactly* two items. #366
* Fixed calls of `CloudLayer.get_parameters()` resulting in errors. #309
* Fixed `SimplexNoiseAlpha` and `FrequencyNoiseAlpha` not handling
  `sigmoid` argument correctly. #343
* Fixed `SnowflakesLayer` crashing for grayscale images. #345
* Fixed `Affine` heatmap augmentation crashing for arrays with more than
  four channels and `order!=0`. #381
* Fixed an outdated error message in `Affine`. #381
* Fixed `Polygon.clip_out_of_image()` crashing if the intersection between
  polygon and image plane was an edge or point. #382
* Fixed `Polygon.clip_out_of_image()` potentially failing for polygons
  containing two or fewer points. #382
* Fixed `Polygon.is_out_of_image()` returning wrong values if the image plane
  was fully contained inside the polygon with no intersection between the
  image plane and the polygon edge. #382
* Fixed  `Fliplr` and `Flipud` using for coordinate-based inputs and image-like
  inputs slightly different conditions for when to actually apply
  augmentations. #385
* Fixed `Convolve` using an overly restrictive check when validating inputs
  for `matrix` w.r.t. whether they are callables. The check should now also
  support class methods (and possibly various other callables). #407
* Fixed `CropAndPad`, `Pad` and `PadToFixedSize` still clipping `cval` samples
  to the `uint8`. They now clip to the input array's dtype's value range. #407
* Fixed `WithColorspace` not propagating polygons to child augmenters. #409
* Fixed `WithHueAndSaturation` not propagating segmentation maps and polygons
  to child augmenters. #409
* Fixed `AlphaElementwise` to blend coordinates (for keypoints, polygons,
  line strings) on a point-by-point basis following the image's average
  alpha value in the sampled alpha mask of the point's coordinate.
  Previously, the average over the whole mask was used and then either all
  points of the first branch or all of the second branch were used as the
  augmentation output. This also affects `SimplexNoiseAlpha` and
  `FrequencyNoiseAlpha`. #410
* Fixed many augmenters and helper functions producing errors if the height,
  width and/or channels of input arrays were exactly `0` or the channels
  were `>512`. #433
* Fixed `Rot90` not supporting `imgaug.ALL`. #434
* Fixed `PiecewiseAffine` possibly generating samples for non-image data
  when using `absolute_scale=True` that were not well aligned with the
  corresponding images. #437
